@@ -29,6 +29,7 @@ public:
     SimApplicationPrivate()
     {
         expectedType = QSimCommand::NoCommand;
+        modemHandled = false;
         target = 0;
         slot = 0;
         inResponse = false;
@@ -36,6 +37,7 @@ public:
 
     SimRules   *rules;
     QSimCommand::Type expectedType;
+    bool modemHandled;
     QByteArray  currentCommand;
     QObject    *target;
     const char *slot;
@@ -86,6 +88,20 @@ void SimApplication::command( const QSimCommand& cmd,
     }
 }
 
+void SimApplication::modemHandledCommand( const QSimCommand& cmd, int timeout)
+{
+    // Record the command details, together with the type of
+    // TERMINAL RESPONSE or ENVELOPE that we expect in answer.
+    d->currentCommand = cmd.toPdu( QSimCommand::NoPduOptions );
+    d->expectedType = cmd.type();
+    d->modemHandled = true;
+
+    if ( d->rules && !d->inResponse ) {
+        d->rules->modemHandledCommandNotify( d->currentCommand );
+    }
+
+    QTimer::singleShot( timeout, this, SLOT(endSession()) );
+}
 /*!
     Sends a call control \a event to the ME.
 */
@@ -118,6 +134,7 @@ void SimApplication::start()
 void SimApplication::abort()
 {
     d->expectedType = QSimCommand::NoCommand;
+    d->modemHandled = false;
     d->currentCommand = QByteArray();
     d->target = 0;
     d->slot = 0;
@@ -162,8 +179,10 @@ bool SimApplication::envelope( const QSimEnvelope& env )
 {
     /* Process a menu selection ENVELOPE message.  We turn it into a
      * QSimTerminalResponse to make it easier to process.  */
-    if ( env.type() == QSimEnvelope::EventDownload )
+    if ( env.type() == QSimEnvelope::EventDownload ) {
+        d->rules->respond( "OK" );
         return true;
+    }
 
     if ( env.type() != QSimEnvelope::MenuSelection )
         /* Not supported */
@@ -173,6 +192,8 @@ bool SimApplication::envelope( const QSimEnvelope& env )
         /* Envelope sent for the wrong type of command. */
         return false;
 
+    d->rules->respond( "OK" );
+
     d->expectedType = QSimCommand::NoCommand;
     d->currentCommand = QByteArray();
     d->target = 0;
@@ -201,6 +222,9 @@ bool SimApplication::response( const QSimTerminalResponse& resp )
          resp.command().type() != d->expectedType )
         return false;
 
+    if (d->modemHandled == true)
+        return false;
+
     // Save the target information.
     QObject *target = d->target;
     const char *slot = d->slot;
@@ -208,6 +232,7 @@ bool SimApplication::response( const QSimTerminalResponse& resp )
     // Clear the command details, in preparation for a new command.
     if ( resp.command().type() != QSimCommand::SetupMenu ) {
         d->expectedType = QSimCommand::NoCommand;
+        d->modemHandled = false;
         d->currentCommand = QByteArray();
     }
     d->target = 0;
@@ -232,22 +257,19 @@ bool SimApplication::response( const QSimTerminalResponse& resp )
 
     // Answer the AT+CSIM command and send notification of the new command.
     if ( !d->rules )
-        return false;////
-    if ( d->currentCommand.isEmpty() || resp.command().type() == QSimCommand::SetupMenu ) {
-        // No new command, so respond with a simple OK.
-        d->rules->respond( "+CSIM: 4,9000\\n\\nOK" );
-    } else {
-        // There is a new command, so send back 91XX to the TERMINAL RESPONSE
-        // or ENVELOPE request to indicate that we have another command to
-        // be fetched.  Then send the unsolicited "*TCMD" notification.
-        QByteArray data;
-        data += (char)0x91;
-        data += (char)(d->currentCommand.size());
-        d->rules->respond( "+CSIM: " + QString::number( data.size() * 2 ) + "," +
-                           QAtUtils::toHex( data ) + "\\n\\nOK" );
-        d->rules->unsolicited
-            ( "*TCMD: " + QString::number( d->currentCommand.size() ) );
-    }
+        return false;
+
+    d->rules->respond("OK");
+
+    // No new command
+    if ( d->currentCommand.isEmpty() ||
+            resp.command().type() == QSimCommand::SetupMenu )
+        return true;
+
+    if (d->modemHandled)
+        d->rules->modemHandledCommandNotify( d->currentCommand );
+    else
+        d->rules->proactiveCommandNotify( d->currentCommand );
 
     return true;
 }
@@ -255,7 +277,8 @@ bool SimApplication::response( const QSimTerminalResponse& resp )
 void SimApplication::endSession()
 {
     d->expectedType = QSimCommand::SetupMenu;
-    d->rules->unsolicited("*TEND");
+    d->modemHandled = false;
+    d->rules->respond( "+CUSATEND", 1);
 }
 
 DemoSimApplication::DemoSimApplication( SimRules *rules, QObject *parent )
@@ -292,6 +315,7 @@ const QString DemoSimApplication::getName()
 #define MainMenu_Refresh    17
 #define MainMenu_LocalInfo  18
 #define MainMenu_BIP        19
+#define MainMenu_Handled    20
 
 #define SportsMenu_Chess        1
 #define SportsMenu_Painting     2
@@ -350,11 +374,15 @@ const QString DemoSimApplication::getName()
 #define LocalInfoMenu_Time  1
 #define LocalInfoMenu_Lang  2
 
-#define BIPMenu_OpenChannel      1
-#define BIPMenu_CloseChannel     2
-#define BIPMenu_ReceiveData      3
-#define BIPMenu_SendData         4
-#define BIPMenu_GetChannelStatus 5
+#define BIPMenu_OpenChannel_TCP        1
+#define BIPMenu_OpenChannel_UDP        2
+#define BIPMenu_CloseChannel           3
+#define BIPMenu_ReceiveData            4
+#define BIPMenu_SendData_Immediately   5
+#define BIPMenu_SendData_Store         6
+#define BIPMenu_GetChannelStatus       7
+
+#define Handled_SendSMS	1
 
 enum SendSMSMenuItems {
 	SendSMS_Unpacked = 1,
@@ -382,6 +410,7 @@ void DemoSimApplication::mainMenu()
     QList<QSimMenuItem> items;
 
     cmd.setType( QSimCommand::SetupMenu );
+    cmd.setTitle( "Phonesim services" );
 
     item.setIdentifier( MainMenu_News );
     item.setLabel( "News" );
@@ -459,6 +488,10 @@ void DemoSimApplication::mainMenu()
     item.setLabel( "BIP commands" );
     items += item;
 
+    item.setIdentifier( MainMenu_Handled );
+    item.setLabel( "Modem-Handled commands" );
+    items += item;
+
     cmd.setMenuItems( items );
 
     command( cmd, 0, 0 );
@@ -607,6 +640,12 @@ void DemoSimApplication::mainMenuSelection( int id )
         }
         break;
 
+        case MainMenu_Handled:
+        {
+            sendHandledMenu();
+        }
+        break;
+
         default:
         {
             // Don't know what this item is, so just re-display the main menu.
@@ -917,6 +956,7 @@ void DemoSimApplication::toneMenu( const QSimTerminalResponse& resp )
     if ( resp.result() == QSimTerminalResponse::Success ) {
         // Item selected.
         cmd.setType( QSimCommand::PlayTone );
+        cmd.setText( "Play a standard supervisory tone" );
         cmd.setDestinationDevice( QSimCommand::Earpiece );
         cmd.setTone( (QSimCommand::Tone)( resp.menuItem() ) );
         if ( cmd.tone() == QSimCommand::ToneDial )
@@ -1367,6 +1407,7 @@ void DemoSimApplication::CFMenu( const QSimTerminalResponse& resp )
                 cmd.setType( QSimCommand::SendSS );
                 cmd.setDestinationDevice( QSimCommand::Network );
                 cmd.setNumber( "**62*+155543*11#" );
+                cmd.setText( "Sending SS Registration" );
                 command( cmd, this, SLOT(sendCFMenu()) );
             }
             break;
@@ -1376,6 +1417,7 @@ void DemoSimApplication::CFMenu( const QSimTerminalResponse& resp )
                 cmd.setType( QSimCommand::SendSS );
                 cmd.setDestinationDevice( QSimCommand::Network );
                 cmd.setNumber( "*62#" );
+                cmd.setText( "Sending SS Activation" );
                 command( cmd, this, SLOT(sendCFMenu()) );
             }
             break;
@@ -1385,6 +1427,7 @@ void DemoSimApplication::CFMenu( const QSimTerminalResponse& resp )
                 cmd.setType( QSimCommand::SendSS );
                 cmd.setDestinationDevice( QSimCommand::Network );
                 cmd.setNumber( "*#62**11#" );
+                cmd.setText( "Sending SS Interrogation class voice" );
                 command( cmd, this, SLOT(sendCFMenu()) );
             }
             break;
@@ -1394,6 +1437,7 @@ void DemoSimApplication::CFMenu( const QSimTerminalResponse& resp )
                 cmd.setType( QSimCommand::SendSS );
                 cmd.setDestinationDevice( QSimCommand::Network );
                 cmd.setNumber( "#62#" );
+                cmd.setText( "Sending SS Deactivation" );
                 command( cmd, this, SLOT(sendCFMenu()) );
             }
             break;
@@ -1403,6 +1447,7 @@ void DemoSimApplication::CFMenu( const QSimTerminalResponse& resp )
                 cmd.setType( QSimCommand::SendSS );
                 cmd.setDestinationDevice( QSimCommand::Network );
                 cmd.setNumber( "##62#" );
+                cmd.setText( "Sending SS Erasure" );
                 command( cmd, this, SLOT(sendCFMenu()) );
             }
             break;
@@ -2376,7 +2421,7 @@ void DemoSimApplication::BIPMenu( const QSimTerminalResponse& resp )
     if ( resp.result() == QSimTerminalResponse::Success ) {
 	switch ( resp.menuItem() ) {
 
-            case BIPMenu_OpenChannel:
+            case BIPMenu_OpenChannel_TCP:
             {
                 QByteArray bearerDesc;
                 QByteArray uti;
@@ -2391,14 +2436,14 @@ void DemoSimApplication::BIPMenu( const QSimTerminalResponse& resp )
                 bearerDesc += 0x1F;
                 bearerDesc += 0x02;
 
-                uti += 0x01;
-                uti += 0xAD;
-                uti += 0x9C;
+                uti += 0x02; /* TCP, UICC in client mode, remote connection */
+                uti += 0x30;
+                uti += 0x3C; /* port 12348 */
 
                 destAddress += 0x21;
-                destAddress += 0x01;
-                destAddress += 0x01;
-                destAddress += 0x01;
+                destAddress += 0x7F; /* local host 127.0.0.1 */
+                destAddress += (char) 0x00;
+                destAddress += (char) 0x00;
                 destAddress += 0x01;
 
                 apn += 0x06;
@@ -2414,7 +2459,57 @@ void DemoSimApplication::BIPMenu( const QSimTerminalResponse& resp )
 
                 cmd.setType( QSimCommand::OpenChannel );
                 cmd.setQualifier( QSimCommand::OpenChannelImmediate );
-                cmd.setText("Open ID");
+                cmd.setText("Open channel TCP");
+                cmd.setBearerDesc(bearerDesc);
+                cmd.setBufferSize(1400);
+                cmd.setApn(apn);
+                cmd.setUserLogin("UserLog");
+                cmd.setUserPassword("UserPwd");
+                cmd.setUti(uti);
+                cmd.setDestAddress(destAddress);
+                command( cmd, this, SLOT(sendBIPMenu()) );
+            }
+            break;
+
+            case BIPMenu_OpenChannel_UDP:
+            {
+                QByteArray bearerDesc;
+                QByteArray uti;
+                QByteArray destAddress;
+                QByteArray apn;
+
+                bearerDesc += 0x02;
+                bearerDesc += 0x03;
+                bearerDesc += 0x04;
+                bearerDesc += 0x03;
+                bearerDesc += 0x04;
+                bearerDesc += 0x1F;
+                bearerDesc += 0x02;
+
+                uti += 0x01; /* UDP, UICC in client mode, remote connection */
+                uti += 0x30;
+                uti += 0x3C; /* port 12348 */
+
+                destAddress += 0x21;
+                destAddress += 0x7F; /* local host 127.0.0.1 */
+                destAddress += (char) 0x00;
+                destAddress += (char) 0x00;
+                destAddress += 0x01;
+
+                apn += 0x06;
+                apn += 0x54;
+                apn += 0x65;
+                apn += 0x73;
+                apn += 0x74;
+                apn += 0x47;
+                apn += 0x70;
+                apn += 0x02;
+                apn += 0x72;
+                apn += 0x73;
+
+                cmd.setType( QSimCommand::OpenChannel );
+                cmd.setQualifier( QSimCommand::OpenChannelImmediate );
+                cmd.setText("Open channel UDP");
                 cmd.setBearerDesc(bearerDesc);
                 cmd.setBufferSize(1400);
                 cmd.setApn(apn);
@@ -2436,19 +2531,39 @@ void DemoSimApplication::BIPMenu( const QSimTerminalResponse& resp )
             }
             break;
 
-            case BIPMenu_SendData:
+            case BIPMenu_SendData_Immediately:
             {
                 QByteArray data;
 
-                data += 0x01;
-                data += 0x02;
-                data += 0x03;
-                data += 0x04;
-                data += 0x05;
+                data += 0x61;
+                data += 0x62;
+                data += 0x63;
+                data += 0x64;
+                data += 0x65;
 
+                cmd.setQualifier( QSimCommand::SendDataImmediately );
                 cmd.setType( QSimCommand::SendData );
                 cmd.setDestinationDevice( QSimCommand::Channel1 );
-                cmd.setText("Send Data 1");
+                cmd.setText("Send Data Immediately");
+                cmd.addExtensionField( 0x36, data );
+                command( cmd, this, SLOT(sendBIPMenu()) );
+            }
+            break;
+
+            case BIPMenu_SendData_Store:
+            {
+                QByteArray data;
+
+                data += 0x31;
+                data += 0x32;
+                data += 0x33;
+                data += 0x34;
+                data += 0x35;
+
+                cmd.setQualifier( QSimCommand::SendDataStoreData );
+                cmd.setType( QSimCommand::SendData );
+                cmd.setDestinationDevice( QSimCommand::Channel1 );
+                cmd.setText("Store Data");
                 cmd.addExtensionField( 0x36, data );
                 command( cmd, this, SLOT(sendBIPMenu()) );
             }
@@ -2488,8 +2603,12 @@ void DemoSimApplication::sendBIPMenu()
     cmd.setType( QSimCommand::SelectItem );
     cmd.setTitle( "BIP commands Menu" );
 
-    item.setIdentifier( BIPMenu_OpenChannel );
-    item.setLabel( "Open channel" );
+    item.setIdentifier( BIPMenu_OpenChannel_TCP );
+    item.setLabel( "Open channel - TCP in client mode" );
+    items += item;
+
+    item.setIdentifier( BIPMenu_OpenChannel_UDP );
+    item.setLabel( "Open channel - UDP in client mode" );
     items += item;
 
     item.setIdentifier( BIPMenu_CloseChannel );
@@ -2500,8 +2619,12 @@ void DemoSimApplication::sendBIPMenu()
     item.setLabel( "Receive data" );
     items += item;
 
-    item.setIdentifier( BIPMenu_SendData );
-    item.setLabel( "Send data" );
+    item.setIdentifier( BIPMenu_SendData_Immediately );
+    item.setLabel( "Send data - Immediately" );
+    items += item;
+
+    item.setIdentifier( BIPMenu_SendData_Store );
+    item.setLabel( "Send data - Store" );
     items += item;
 
     item.setIdentifier( BIPMenu_GetChannelStatus );
@@ -2512,3 +2635,62 @@ void DemoSimApplication::sendBIPMenu()
 
     command( cmd, this, SLOT(BIPMenu(QSimTerminalResponse)) );
 }
+
+void DemoSimApplication::sendHandledMenu()
+{
+    QSimCommand cmd;
+    QSimMenuItem item;
+    QList<QSimMenuItem> items;
+
+    cmd.setType( QSimCommand::SelectItem );
+    cmd.setTitle( "Modem-Handled Commands Menu" );
+
+    item.setIdentifier( Handled_SendSMS );
+    item.setLabel( "Send Modem-Handled SMS" );
+    items += item;
+
+    cmd.setMenuItems( items );
+
+    command( cmd, this, SLOT(handledMenuResp(QSimTerminalResponse)) );
+}
+
+void DemoSimApplication::handledMenuResp( const QSimTerminalResponse& resp )
+{
+    QSimCommand cmd;
+
+    if ( resp.result() != QSimTerminalResponse::Success ) {
+        /* Unknown response - just go back to the main menu. */
+        endSession();
+
+        return;
+    }
+
+    /* Item selected. */
+    switch ( resp.menuItem() ) {
+    case Handled_SendSMS:
+    {
+        QSMSMessage sms;
+        sms.setValidityPeriod( -1 );
+        sms.setMessageClass( 2 );
+        sms.setProtocol( 0 );
+        sms.setRecipient( "12345" );
+        sms.setText( smsText );
+        sms.setForceGsm( false );
+        sms.setBestScheme( QSMS_8BitAlphabet );
+        sms.setDataCodingScheme( 0xf6 );
+
+        cmd.setType( QSimCommand::SendSMS );
+        cmd.setText( "Sending an SMS to our friends at 12345" );
+        cmd.setNumber( "123" );
+        cmd.addExtensionField( 0x8b, sms.toPdu().mid( 1 ) );
+        cmd.setDestinationDevice( QSimCommand::Network );
+
+        modemHandledCommand(cmd, 6000);
+        break;
+    }
+
+    default:
+	endSession();
+	break;
+    }
+}
